#ifndef __BUILTINS_LOADED
#define __BUILTINS_LOADED 1

/*
 *  <builtins.h> - Prototypes for platform specific builtins
 *
 *  Copyright (c) 1993 by Digital Equipment Corporation.  All rights reserved.
 *
 *  DEC C for OpenVMS AXP
 *  DEC C++ for OpenVMS AXP
 */

#if defined(__ALPHA)

#include <ints.h>

#if defined(VMS) || defined(__VMS)
/*
** xxxQUE Mapping Tables
**
** These tables map the _PAL returned values 
** into VAX C expected values.
**
*/

static const int __REMQUE_MAP_ALPHA_TO_VAX[] = {2,1,0};
static const int __INSQUE_MAP_ALPHA_TO_VAX[] = {1,0};
static const int __INSQHI_MAP_ALPHA_TO_VAX[] = {1,0,2};
static const int __INSQTI_MAP_ALPHA_TO_VAX[] = {1,0,2};
static const int __REMQHI_MAP_ALPHA_TO_VAX[] = {1,3,0,2};
static const int __REMQTI_MAP_ALPHA_TO_VAX[] = {1,3,0,2};
static const int __PAD_BUFFER[] = {0};
#endif  /* VMS || __VMS */

/*
** __ALLOCA builtin - allocate n-bytes from the stack
*/

void * __ALLOCA(unsigned int x);

/*
** UMULH Builtin - Unsigned Quadword Multiply High
*/
uint64 __UMULH(uint64 operand1, uint64 operand2);

/*
** op_ATOMIC_size Builtins
**
** Note: There is one optional parameter.
**
**	 - an optional input for a retry count.
*/    
int __ADD_ATOMIC_LONG(void *address, int expression, ...);
int __ADD_ATOMIC_QUAD(void *address, int expression, ...);

int __AND_ATOMIC_LONG(void *address, int expression, ...);
int __AND_ATOMIC_QUAD(void *address, int expression, ...);

int __OR_ATOMIC_LONG(void *address, int expression, ...);
int __OR_ATOMIC_QUAD(void *address, int expression, ...);

/*
** TESTBITxxI
**  
** Note: There is one optional parameter.
**
**	 - an optional input for a retry count.
**
*/
int __TESTBITCCI(void *address, int position, ...);
int __TESTBITSSI(void *address, int position, ...);


/*
** Add Aligned Word Interlocked
*/
int __ADAWI(short src, short *dest);

/*
** Trap Barrier Instruction
*/
void __TRAPB(void);

/*
** Read Cycle Counter
*/
uint64 __RPCC(void);

/*
** Halt the Processor. (Privileged)
*/
void _HALT(void);
void __PAL_HALT(void);


/*
** Generate Trap
*/
void __PAL_GENTRAP(uint64 encoded_software_trap);

#if defined(VMS) || defined(__VMS)
/* 
** Insert entry into longword queue
*/
int __PAL_INSQHIL(void *head, void *new_entry);	      /* At head, interlocked */
int __PAL_INSQTIL(void *head, void *new_entry);	      /* At tail, interlocked */
int __PAL_INSQUEL(void *predecessor, void *new_entry); 
int __PAL_INSQUEL_D(void **predecessor, void *new_entry); /* Deferred */

/* 
** Insert entry into quadword queue
*/
int __PAL_INSQHIQ(void *head, void *new_entry);	      /* At head, interlocked */
int __PAL_INSQTIQ(void *head, void *new_entry);	      /* At tail, interlocked */
int __PAL_INSQUEQ(void *predecessor, void *new_entry);
int __PAL_INSQUEQ_D(void **predecessor, void *new_entry); /* Deferred */

/*
** Insert Resident Interlocked Queue Opcodes
*/
int __PAL_INSQHILR(void *head, void *new_entry);      /* At head, interlocked */
int __PAL_INSQTILR(void *head, void *new_entry);      /* At tail, interlocked */
int __PAL_INSQHIQR(void *head, void *new_entry);      /* At head, interlocked */
int __PAL_INSQTIQR(void *head, void *new_entry);      /* At tail, interlocked */

/* 
** Remove entry from longword queue
*/
int __PAL_REMQHIL(void *head, void **removed_entry);	/* At head, interlocked */
int __PAL_REMQTIL(void *head, void **removed_entry);	/* At tail, interlocked */
int __PAL_REMQUEL(void *entry, void **removed_entry);
int __PAL_REMQUEL_D(void **entry, void **removed_entry); /* Deferred */

/* 
** Remove entry from quadword queue
*/
int __PAL_REMQHIQ(void *head, void **removed_entry);     /* At head, interlocked */
int __PAL_REMQTIQ(void *head, void **removed_entry);     /* At tail, interlocked */
int __PAL_REMQUEQ(void *entry, void **removed_entry);   
int __PAL_REMQUEQ_D(void **entry, void **removed_entry); /* Deferred */

/*
** Remove Resident Interlocked Queue Opcodes
*/
int __PAL_REMQHILR(void *head, void **removed_entry);	/* At head, interlocked */
int __PAL_REMQTILR(void *head, void **removed_entry);    /* At tail, interlocked */
int __PAL_REMQHIQR(void *head, void **removed_entry);    /* At head, interlocked */
int __PAL_REMQTIQR(void *head, void **removed_entry);    /* At tail, interlocked */


/*
** Move from Processor Register
*/
unsigned int __PAL_MFPR_ASTEN(void);   /* AST Enable */
unsigned int __PAL_MFPR_ASTSR(void);   /* AST Summary Register */
void *__PAL_MFPR_ESP(void);	       /* Executive Stack Pointer */
int __PAL_MFPR_FEN(void);	       /* Floating Point Enable */
int __PAL_MFPR_IPL(void);	       /* Interrupt Priority Level */
int __PAL_MFPR_MCES(void);	       /* Machine Check Error Summary */
void *__PAL_MFPR_PCBB(void);	       /* Privileged Context Block Base */
int64 __PAL_MFPR_PRBR(void);	       /* Processor Base Register */
int __PAL_MFPR_PTBR(void);	       /* Page Table Base Register */
void *__PAL_MFPR_SCBB(void);	       /* System Control Block Base */
unsigned int __PAL_MFPR_SISR(void);    /* Software Interrupt Summary Register */
void *__PAL_MFPR_SSP(void);	       /* Supervisor Stack Pointer */
int64 __PAL_MFPR_TBCHK(void *address); /* Translation Buffer Check */
void *__PAL_MFPR_USP(void);	       /* User Stack Pointer */
void *__PAL_MFPR_VPTB(void);	       /* Virtual Page Table */
int64 __PAL_MFPR_WHAMI(void);	       /* Who Am I */

/*
** Move to Processor Register
*/
void __PAL_MTPR_ASTEN(unsigned int mask);    /* AST Enable */
void __PAL_MTPR_ASTSR(unsigned int mask);    /* AST Summary Register */
void __PAL_MTPR_DATFX(int value);	     /* Data Alignment Trap Fixup */
void __PAL_MTPR_ESP(void *address);	     /* Executive Stack Pointer */
void __PAL_MTPR_FEN(int value);		     /* Floating Point Enable */
void __PAL_MTPR_IPIR(int64 number);	     /* Interprocessor Interrupt Request */
int __PAL_MTPR_IPL(int value);		     /* Interrupt Priority Level */
void __PAL_MTPR_MCES(int value);	     /* Machine Check Error Summary */
void __PAL_MTPR_PRBR(int64 value);	     /* Processor Base Register */
void __PAL_MTPR_SCBB(void *address);	     /* System Control Block Base */
void __PAL_MTPR_SIRR(int level);	     /* Software Interrupt Request Register */
void __PAL_MTPR_SSP(int *address);	     /* Supervisor Stack Pointer */
void __PAL_MTPR_TBIA(void);		     /* User Stack Pointer */
void __PAL_MTPR_TBIAP(void);		     /* Translation Buffer Invalidate All Process */
void __PAL_MTPR_TBIS(void *address);	     /* Translation Buffer Invalidate Single */
void __PAL_MTPR_TBISD(void *address);	     /* Translation Buffer Invalidate Single Data */
void __PAL_MTPR_TBISI(void *address);	     /* Translation Buffer Invalidate Single Instruction */
void __PAL_MTPR_USP(void *address);	     /* User Stack Pointer */
void __PAL_MTPR_VPTB(void *address);	     /* Virtual Page Table */

/*
** Probe Read/Write Acessibility
*/
int __PAL_PROBER(const void *base_address, int length, char mode);
int __PAL_PROBEW(const void *base_address, int length, char mode);

/*
** Change Mode 
*/
void __PAL_CHME(void);                           /* Executive  */
void __PAL_CHMK(void);                           /* Kernel     */
void __PAL_CHMS(void);                           /* Supervisor */
void __PAL_CHMU(void);                           /* User       */

/*
** Load/Store Quadword Physical
*/
uint64 __PAL_LDQP(void *address);                /* Load       */
void   __PAL_STQP(void *address, uint64 value);  /* Store      */
#endif  /* VMS || __VMS */

/*
** Breakpoint
*/
void __PAL_BPT(void);

/*
** Bugcheck
*/
void __PAL_BUGCHK(void);

#if defined(VMS) || defined(__VMS)
/*
** Cache Flush
*/
void __PAL_CFLUSH(int value);

/*
** Drain Aborts
*/
void __PAL_DRAINA(void);

/*
** Read Processor Status
*/
uint64 __PAL_RD_PS(void);
#endif  /* VMS || __VMS */

/*
** Swap Privileged Context
*/
void __PAL_SWPCTX(void *address);

#if defined(VMS) || defined(__VMS)
/*
** Swap AST Enable
*/
unsigned int __PAL_SWASTEN(int new_state_mask);

/*
** Write Processor Status Software Field
*/
void __PAL_WR_PS_SW(int mask);
#endif  /* VMS || __VMS */


/*
** Copy Sign
*/
float	    __CPYSF(float operand1, float operand2);
double	    __CPYS(double operand1, double operand2);

/*
** Copy Sign Negate
*/
float	    __CPYSNF(float operand1, float operand2);
double	    __CPYSN(double operand1, double operand2);

/*
** Copy Sign Exponent
*/
float	    __CPYSEF(float operand1, float operand2);
double	    __CPYSE(double operand1, double operand2);

#if defined(VMS) || defined(__VMS)
/*
** Convert from G-Floating to Quadword
*/
int64 __CVTGQ(double operand1);

/*
** Convert from G-Floating to F-Floating Chopped
*/
float __CVTGF_C(double operand1);
#endif  /* VMS || __VMS */

/*
** Convert from T-Floating to Quadword
*/
int64 __CVTTQ(double operand1);

/*
** Convert from T-Floating to S-Floating Chopped
*/
float __CVTTS_C(double operand1);

/*
**  Add Floating Point Chopped
*/
#if defined(VMS) || defined(__VMS)
float __ADDF_C(float operand1, float operand2);
double __ADDG_C(double operand1, double operand2);
#endif  /* VMS || __VMS */
float __ADDS_C(float operand1, float operand2);
double __ADDT_C(double operand1, double operand2);

/*
**  Subtract Floating Point Chopped
*/
#if defined(VMS) || defined(__VMS)
float __SUBF_C(float operand1, float operand2);
double __SUBG_C(double operand1, double operand2);
#endif  /* VMS || __VMS */
float __SUBS_C(float operand1, float operand2);
double __SUBT_C(double operand1, double operand2);

/*
**  Multiply Floating Point Chopped
*/
#if defined(VMS) || defined(__VMS)
float __MULF_C(float operand1, float operand2);
double __MULG_C(double operand1, double operand2);
#endif  /* VMS || __VMS */
float __MULS_C(float operand1, float operand2);
double __MULT_C(double operand1, double operand2);

/*
**  Divide Floating Point Chopped
*/
#if defined(VMS) || defined(__VMS)
float __DIVF_C(float operand1, float operand2);
double __DIVG_C(double operand1, double operand2);
#endif  /* VMS || __VMS */
float __DIVS_C(float operand1, float operand2);
double __DIVT_C(double operand1, double operand2);


/*
** Memory Barrier
*/
void	__MB(void);

/*
** Instruction Memory Barrier
*/
void	__PAL_IMB(void);

/*
** Compare, Store Long/Quad
*/
int __CMP_STORE_LONG(
    void *source,
    int  old_value,
    int  new_value,
    void *dest);
int __CMP_STORE_QUAD(
    void  *source,
    int64 old_value,
    int64 new_value,
    void  *dest);

/*
** Macros for translation from VAX C to DEC C ALPHA builtins
*/
#if defined(VMS) || defined(__VMS)

#define _BBCCI(position, address)			\
	__TESTBITCCI((address), (position))

#define _BBSSI(position, address)			\
	__TESTBITSSI((address), (position))

#define _INSQHI(new_entry, head) 			\
        __INSQHI_MAP_ALPHA_TO_VAX[__PAL_INSQHIL((head), (new_entry))+1]

#define _INSQTI(new_entry, head) 			\
        __INSQTI_MAP_ALPHA_TO_VAX[__PAL_INSQTIL((head), (new_entry))+1]

#define _INSQUE(new_entry, predecessor)	 		\
        __INSQUE_MAP_ALPHA_TO_VAX[__PAL_INSQUEL((predecessor), (new_entry))]

#define _REMQHI(head, removed_entry) 			\
	__REMQHI_MAP_ALPHA_TO_VAX[__PAL_REMQHIL((head), (removed_entry))+1]

#define _REMQTI(head, removed_entry)	 		\
	__REMQTI_MAP_ALPHA_TO_VAX[__PAL_REMQTIL((head), (removed_entry))+1]

#define _REMQUE(entry, removed_entry) 			\
	__REMQUE_MAP_ALPHA_TO_VAX[__PAL_REMQUEL((entry), (removed_entry))+1]

#define _PROBER(mode, length, address)			\
	__PAL_PROBER((address), (length), (mode))

#define _PROBEW(mode, length, address)			\
	__PAL_PROBEW((address), (length), (mode))

#endif    /* VMS || __VMS */
#endif   /* __ALPHA */
#endif   /* __BUILTINS_LOADED */
